!*****************************************************************************

!   DSFUN: A Double-Single Floating Point Computation Package

!   IEEE Fortran-90 version
!   Version Date:  2005-01-26

!   Author:

!      David H. Bailey
!      NERSC, Lawrence Berkeley Lab
!      Mail Stop 50B-2239
!      Berkeley, CA 94720
!      Email: dhbailey@lbl.gov

!   Some system-dependent items are noted with !> comments.

!*****************************************************************************

module dsfunmod

contains

subroutine dsabrt
implicit none

!   This permits one to insert a call to a vendor-specific traceback routine.

stop
end subroutine

subroutine dsadd (dsa, dsb, dsc)

!   This subroutine computes dsc = dsa + dsb.

implicit none
real*4 dsa(2), dsb(2), dsc(2)
real*4 e, t1, t2

!   Compute dsa + dsb using Knuth's trick.

t1 = dsa(1) + dsb(1)
e = t1 - dsa(1)
t2 = ((dsb(1) - e) + (dsa(1) - (t1 - e))) + dsa(2) + dsb(2)

!   The result is t1 + t2, after normalization.

dsc(1) = t1 + t2
dsc(2) = t2 - (dsc(1) - t1)
return
end subroutine

subroutine dsang (x, y, a)

!   This computes the DS angle A subtended by the DS pair (X, Y) considered as
!   a point in the x-y plane.  This is more useful than an arctan or arcsin
!   routine, since it places the result correctly in the full circle, i.e.
!   -Pi < A <= Pi.

!   The Taylor series for Sin converges much more slowly than that of Arcsin.
!   Thus this routine does not employ Taylor series, but instead computes
!   Arccos or Arcsin by solving Cos (a) = x or Sin (a) = y using one of the
!   following Newton iterations, both of which converge to a:

!           z_{k+1} = z_k - [x - Cos (z_k)] / Sin (z_k)
!           z_{k+1} = z_k + [y - Sin (z_k)] / Cos (z_k)

!   The first is selected if Abs (x) <= Abs (y); otherwise the second is used.

implicit none
integer i, ix, iy, k, kk, nx, ny
real*4 t1, t2, t3
real*4 a(2), pi(2), x(2), y(2), s0(2), s1(2), s2(2), s3(2), s4(2)
save pi
data pi/ 3.14159274E+00, -8.74227766E-08/

!   Check if both X and Y are zero.

if (x(1) .eq. 0. .and. y(1) .eq. 0.) then
  write (6, 1)
1 format ('*** DSANG: Both arguments are zero.')
  call dsabrt
  return
endif

!   Check if one of X or Y is zero.

if (x(1) .eq. 0.) then
  if (y(1) .gt. 0.) then
    call dsmuls (pi, 0.5, a)
  else
    call dsmuls (pi, -0.5, a)
  endif
  goto 120
elseif (y(1) .eq. 0.) then
  if (x(1) .gt. 0.) then
      a(1) = 0.
      a(2) = 0.
  else
    a(1) = pi(1)
    a(2) = pi(2)
  endif
  goto 120
endif

!   Normalize x and y so that x^2 + y^2 = 1.

call dsmul (x, x, s0)
call dsmul (y, y, s1)
call dsadd (s0, s1, s2)
call dssqrt (s2, s3)
call dsdiv (x, s3, s1)
call dsdiv (y, s3, s2)

!   Compute initial approximation of the angle.

call dsqdc (s1, t1)
call dsqdc (s2, t2)
t3 = atan2 (t2, t1)
a(1) = t3
a(2) = 0.

!   The smaller of x or y will be used from now on to measure convergence.
!   This selects the Newton iteration (of the two listed above) that has the
!   largest denominator.

if (abs (t1) .le. abs (t2)) then
  kk = 1
  s0(1) = s1(1)
  s0(2) = s1(2)
else
  kk = 2
  s0(1) = s2(1)
  s0(2) = s2(2)
endif

!   Perform the Newton-Raphson iteration described.

do k = 1, 3
  call dscssn (a, s1, s2)
  if (kk .eq. 1) then
    call dssub (s0, s1, s3)
    call dsdiv (s3, s2, s4)
    call dssub (a, s4, s1)
  else
    call dssub (s0, s2, s3)
    call dsdiv (s3, s1, s4)
    call dsadd (a, s4, s1)
  endif
  a(1) = s1(1)
  a(2) = s1(2)
enddo

 120  continue

return
end subroutine

subroutine dscpr (a, b, ic)

!   This routine compares the DS numbers A and B and returns in IC the value
!   -1, 0, or 1 depending on whether A < B, A = B, or A > B.  It is faster
!   than merely subtracting A and B and looking at the sign of the result.

implicit none
integer ic
real*4 a(2), b(2)

if (a(1) .lt. b(1)) then
  ic = -1
elseif (a(1) .eq. b(1)) then
  if (a(2) .lt. b(2)) then
    ic = -1
  elseif (a(2) .eq. b(2)) then
    ic = 0
  else
    ic = 1
  endif
else
  ic = 1
endif

return
end subroutine

subroutine dscssh (a, x, y)

!   This computes the hyperbolic cosine and sine of the DS number A and
!   returns the two DS results in X and Y, respectively. 

implicit none
real*4 a(2), f(2), x(2), y(2), s0(2), s1(2), s2(2)

f(1) = 1.
f(2) = 0.
call dsexp (a, s0)
call dsdiv (f, s0, s1)
call dsadd (s0, s1, s2)
call dsmuls (s2, 0.5, x)
call dssub (s0, s1, s2)
call dsmuls (s2, 0.5, y)

return
end subroutine

subroutine dscssn (a, x, y)

!   This computes the cosine and sine of the DS number A and returns the two DS
!   results in X and Y, respectively.

!   This routine uses the conventional Taylor's series for Sin (s):

!   Sin (s) =  s - s^3 / 3! + s^5 / 5! - s^7 / 7! ...

!   where s = t - a * pi / 2 - b * pi / 16 and the integers a and b are chosen
!   to minimize the absolute value of s.  We can then compute

!   Sin (t) = Sin (s + a * pi / 2 + b * pi / 16)
!   Cos (t) = Cos (s + a * pi / 2 + b * pi / 16)

!   by applying elementary trig identities for sums.  The sine and cosine of
!   b * pi / 16 are of the form 1/2 * Sqrt {2 +- Sqrt [2 +- Sqrt(2)]}.
!   Reducing t in this manner insures that -Pi / 32 < s <= Pi / 32, which
!   accelerates convergence in the above series.

implicit none
integer ia, ka, kb, kc, l1
real*4 t1, t2
real*4 a(2), f(2), pi(2), x(2), y(2), s0(2), s1(2), s2(2), s3(2), s4(2), &
  s5(2), s6(2)
real*4 cs(2,2,4)
save cs, pi
data cs/ &
   9.80785251E-01,  2.97394731E-08,  1.95090324E-01, -1.67047143E-09, &
   9.23879504E-01,  2.83074897E-08,  3.82683426E-01,  6.22335072E-09, &
   8.31469595E-01,  1.68702634E-08,  5.55570245E-01, -1.17695214E-08, &
   7.07106769E-01,  1.21016175E-08,  7.07106769E-01,  1.21016175E-08/
data pi/ 3.14159274E+00, -8.74227766E-08/

if (a(1) .eq. 0.) then
  x(1) = 1.
  x(2) = 0.
  y(1) = 0.
  y(2) = 0.
  goto 120
endif

f(1) = 1.
f(2) = 0.

!   Reduce to between - Pi and Pi.

call dsmuls (pi, 2., s0)
call dsdiv (a, s0, s1)
call dsnint (s1, s2)
call dssub (s1, s2, s3)

!   Determine nearest multiple of Pi / 2, and within a quadrant, the nearest
!   multiple of Pi / 16.  Through most of the rest of this subroutine, KA and
!   KB are the integers a and b of the algorithm above.

t1 = s3(1)
t2 = 4. * t1
ka = nint (t2)
kb = nint (8. * (t2 - ka))
t1 = (8 * ka + kb) / 32.
s1(1) = t1
s1(2) = 0.
call dssub (s3, s1, s2)
call dsmul (s0, s2, s1)

!   Compute cosine and sine of the reduced argument s using Taylor's series.

if (s1(1) .eq. 0.) then
  s0(1) = 0.
  s0(2) = 0.
  goto 110
endif
s0(1) = s1(1)
s0(2) = s1(2)
call dsmul (s0, s0, s2)
l1 = 0

100  l1 = l1 + 1
if (l1 .eq. 100) then
  write (6, 1)
1 format ('*** DSCSSN: Iteration limit exceeded.')
  call dsabrt
  return
endif

t2 = - (2. * l1) * (2. * l1 + 1.)
call dsmul (s2, s1, s3)
call dsdivs (s3, t2, s1)
call dsadd (s1, s0, s3)
s0(1) = s3(1)
s0(2) = s3(2)

!   Check for convergence of the series.

if (abs (s1(1)) .gt. 1e-17 * abs (s3(1))) goto 100

!   Compute Cos (s) = Sqrt [1 - Sin^2 (s)].

110  continue

s1(1) = s0(1)
s1(2) = s0(2)
call dsmul (s0, s0, s2)
call dssub (f, s2, s3)
call dssqrt (s3, s0)

!   Compute cosine and sine of b * Pi / 16.

kc = abs (kb)
f(1) = 2.
if (kc .eq. 0) then
  s2(1) = 1.
  s2(2) = 0.
  s3(1) = 0.
  s3(2) = 0.
else
  s2(1) = cs(1,1,kc)
  s2(2) = cs(2,1,kc)
  s3(1) = cs(1,2,kc)
  s3(2) = cs(2,2,kc)
endif
if (kb .lt. 0) then
  s3(1) = - s3(1)
  s3(2) = - s3(2)
endif

!   Apply the trigonometric summation identities to compute cosine and sine of
!   s + b * Pi / 16.

call dsmul (s0, s2, s4)
call dsmul (s1, s3, s5)
call dssub (s4, s5, s6)
call dsmul (s1, s2, s4)
call dsmul (s0, s3, s5)
call dsadd (s4, s5, s1)
s0(1) = s6(1)
s0(2) = s6(2)

!   This code in effect applies the trigonometric summation identities for
!   (s + b * Pi / 16) + a * Pi / 2.

if (ka .eq. 0) then
  x(1) = s0(1)
  x(2) = s0(2)
  y(1) = s1(1)
  y(2) = s1(2)
elseif (ka .eq. 1) then
  x(1) = - s1(1)
  x(2) = - s1(2)
  y(1) = s0(1)
  y(2) = s0(2)
elseif (ka .eq. -1) then
  x(1) = s1(1)
  x(2) = s1(2)
  y(1) = - s0(1)
  y(2) = - s0(2)
elseif (ka .eq. 2 .or. ka .eq. -2) then
  x(1) = - s0(1)
  x(2) = - s0(2)
  y(1) = - s1(1)
  y(2) = - s1(2)
endif

120  continue

return
end subroutine

subroutine dsdiv (dsa, dsb, dsc)

!   This divides the DS number DSA by the DS number DSB to yield the DS
!   quotient DSC.

implicit none
real*4 dsa(2), dsb(2), dsc(2)
real*4 a1, a2, b1, b2, cona, conb, c11, c2, c21, e, split, s1, s2, &
  t1, t2, t11, t12, t21, t22
parameter (split = 8193.)

!   Compute a DP approximation to the quotient.

s1 = dsa(1) / dsb(1)
!>
!   On systems with a fused multiply add, such as IBM systems, it is faster to
!   uncomment the next two lines and comment out the following lines until !>.
!   On other systems, do the opposite.

! c11 = s1 * dsb(1)
! c21 = s1 * dsb(1) - c11

!   This splits s1 and dsb(1) into high-order and low-order words.

cona = s1 * split
conb = dsb(1) * split
a1 = cona - (cona - s1)
b1 = conb - (conb - dsb(1))
a2 = s1 - a1
b2 = dsb(1) - b1

!   Multiply s1 * dsb(1) using Dekker's method.

c11 = s1 * dsb(1)
c21 = (((a1 * b1 - c11) + a1 * b2) + a2 * b1) + a2 * b2
!>
!   Compute s1 * dsb(2) (only high-order word is needed).

c2 = s1 * dsb(2)

!   Compute (c11, c21) + c2 using Knuth's trick.

t1 = c11 + c2
e = t1 - c11
t2 = ((c2 - e) + (c11 - (t1 - e))) + c21

!   The result is t1 + t2, after normalization.

t12 = t1 + t2
t22 = t2 - (t12 - t1)

!   Compute dsa - (t12, t22) using Knuth's trick.

t11 = dsa(1) - t12
e = t11 - dsa(1)
t21 = ((-t12 - e) + (dsa(1) - (t11 - e))) + dsa(2) - t22

!   Compute high-order word of (t11, t21) and divide by dsb(1).

s2 = (t11 + t21) / dsb(1)

!   The result is s1 + s2, after normalization.

dsc(1) = s1 + s2
dsc(2) = s2 - (dsc(1) - s1)

return
end subroutine

subroutine dsdivs (dsa, db, dsc)

!   This routine divides the DS number A by the DP number B to yield
!   the DS quotient C.  

implicit none
real*4 dsa(2), db, dsc(2)
real*4 a1, a2, b1, b2, cona, conb, e, split, t1, t2, t11, t12, t21, t22
parameter (split = 8193.)

!   Compute a DP approximation to the quotient.

t1 = dsa(1) / db
!>
!   On systems with a fused multiply add, such as IBM systems, it is faster to
!   uncomment the next two lines and comment out the following lines until !>.
!   On other systems, do the opposite.

! t12 = t1 * db
! t22 = t1 * db - t12

!   This splits t1 and db into high-order and low-order words.

cona = t1 * split
conb = db * split
a1 = cona - (cona - t1)
b1 = conb - (conb - db)
a2 = t1 - a1
b2 = db - b1

!   Multiply t1 * db using Dekker's method.

t12 = t1 * db
t22 = (((a1 * b1 - t12) + a1 * b2) + a2 * b1) + a2 * b2
!>
!   Compute dsa - (t12, t22) using Knuth's trick.

t11 = dsa(1) - t12
e = t11 - dsa(1)
t21 = ((-t12 - e) + (dsa(1) - (t11 - e))) + dsa(2) - t22

!   Compute high-order word of (t11, t21) and divide by db.

t2 = (t11 + t21) / db

!   The result is t1 + t2, after normalization.

dsc(1) = t1 + t2
dsc(2) = t2 - (dsc(1) - t1)
return
end subroutine

subroutine dsdqc (a, b)

!   This routine converts the DP number A to DS form in B.  All bits of
!   A are recovered in B.  However, note for example that if A = 0.1D0 and N
!   is 0, then B will NOT be the DS equivalent of 1/10.

implicit none
real*4 a, b(2)

b(1) = a
b(2) = 0.
return
end subroutine

subroutine dseq (a, b)

!   This routine sets the DS number B equal to the DS number A. 

implicit none
real*4 a(2), b(2)

b(1) = a(1)
b(2) = a(2)
end subroutine

subroutine dsexp (a, b)

!   This computes the exponential function of the DS number A and returns the
!   DS result in B.

!   This routine uses a modification of the Taylor's series for Exp (t):

!   Exp (t) =  (1 + r + r^2 / 2! + r^3 / 3! + r^4 / 4! ...) ^ q * 2 ^ n

!   where q = 64, r = t' / q, t' = t - n Log(2) and where n is chosen so
!   that -0.5 Log(2) < t' <= 0.5 Log(2).  Reducing t mod Log(2) and
!   dividing by 64 insures that -0.004 < r <= 0.004, which accelerates
!   convergence in the above series.

implicit none
integer i, ia, l1, na, nq, nz, n1
real*4 t1, t2
parameter (nq = 6)
real*4 a(2), b(2), al2(2), f(2), s0(2), s1(2), s2(2), s3(2), tl
save al2
data al2/6.93147182E-01, -1.90465421E-09/

!   Check for overflows and underflows.

if (abs (a(1)) .ge. 709.) then
  if (a(1) .gt. 0.) then
    write (6, 1) a(1)
1   format ('*** DSEXP: Argument is too large',f12.6)
    call dsabrt
    return
  else
    call dsdqc (0., b)
    goto 130
  endif
endif

f(1) = 1.
f(2) = 0.

!   Compute the reduced argument A' = A - Log(2) * Nint [A / Log(2)].  Save
!   NZ = Nint [A / Log(2)] for correcting the exponent of the final result.

call dsdiv (a, al2, s0)
call dsnint (s0, s1)
t1 = s1(1)
nz = t1 + sign (1.e-7, t1)
call dsmul (al2, s1, s2)
call dssub (a, s2, s0)

!   Check if the reduced argument is zero.

if (s0(1) .eq. 0.) then
  s0(1) = 1.
  s0(2) = 0.
  l1 = 0
  goto 120
endif

!   Divide the reduced argument by 2 ^ NQ.

call dsdivs (s0, 2. ** nq, s1)

!   Compute Exp using the usual Taylor series.

s2(1) = 1.
s2(2) = 0.
s3(1) = 1.
s3(2) = 0.
l1 = 0

100  l1 = l1 + 1

if (l1 .eq. 100) then
  write (6, 2)
2 format ('*** DSEXP: Iteration limit exceeded.')
  call dsabrt
  return
endif

t2 = l1
call dsmul (s2, s1, s0)
call dsdivs (s0, t2, s2)
call dsadd (s3, s2, s0)
call dseq (s0, s3)

!   Check for convergence of the series.

if (abs (s2(1)) .gt. 1e-16 * abs (s3(1))) goto 100

!   Raise to the (2 ^ NQ)-th power.

do i = 1, nq
  call dsmul (s0, s0, s1)
  s0(1) = s1(1)
  s0(2) = s1(2)
enddo

!  Multiply by 2 ^ NZ.

120  call dsmuls (s0, 2. ** nz, b)

!   Restore original precision level.

 130  continue

return
end subroutine

subroutine dsinfr (a, b, c)

!   Sets B to the integer part of the DS number A and sets C equal to the
!   fractional part of A.  Note that if A = -3.3, then B = -3 and C = -0.3.

implicit none
integer ic
real*4 a(2), b(2), c(2), con(2), f(2), s0(2), s1(2), t47, t23
parameter (t47 = 2. ** 47, t23 = 2. ** 23)
save con
data con / t47, t23/

!   Check if  A  is zero.

if (a(1) .eq. 0.)  then
  b(1) = 0.
  b(2) = 0.
  c(1) = 0.
  c(2) = 0.
  goto 120
endif

if (a(1) .ge. t47) then
  write (6, 1)
1 format ('*** DSINFR: Argument is too large.')
  call dsabrt
  return
endif

f(1) = 1.
f(2) = 0.
if (a(1) .gt. 0.) then
  call dsadd (a, con, s0)
  call dssub (s0, con, b)
  call dscpr (a, b, ic)
  if (ic .ge. 0) then
    call dssub (a, b, c)
  else
    call dssub (b, f, s1)
    b(1) = s1(1)
    b(2) = s1(2)
    call dssub (a, b, c)
  endif
else
  call dssub (a, con, s0)
  call dsadd (s0, con, b)
  call dscpr (a, b, ic)
  if (ic .le. 0) then
    call dssub (a, b, c)
  else
    call dsadd (b, f, s1)
    b(1) = s1(1)
    b(2) = s1(2)
    call dssub (a, b, c)
  endif
endif

120  continue
return
end subroutine

subroutine dsinp (iu, a)

!   This routine reads the DS number A from logical unit IU.  The input
!   value must be placed on a single line of not more than 80 characters.

implicit none
integer iu, ln
parameter (ln = 80)
character*80 cs
real*4 a(2)

read (iu, '(a)', end = 100) cs
call dsinpc (cs, a)
goto 110

100 write (6, 1)
1  format ('*** DSINP: End-of-file encountered.')
call dsabrt

110 return
end subroutine

subroutine dsinpc (a, b)

!   Converts the CHARACTER*80 array A into the DS number B.

implicit none
integer i, ib, id, ie, inz, ip, is, ix, k, ln, lnn
parameter (ln = 80)
real*4 bi
character*80 a
character*1 ai
character*10 ca, dig
parameter (dig = '0123456789')
real*4 b(2), f(2), s0(2), s1(2), s2(2)

id = 0
ip = -1
is = 0
inz = 0
s1(1) = 0.
s1(2) = 0.

do i = 80, 1, -1
  if (a(i:i) /= ' ') goto 90
enddo

90 continue

lnn = i

!   Scan for digits, looking for the period also.

do i = 1, lnn
  ai = a(i:i)
  if (ai .eq. ' ' .and. id == 0) then
  elseif (ai .eq. '.') then
    if (ip >= 0) goto 210
    ip = id
    inz = 1
  elseif (ai .eq. '+') then
    if (id .ne. 0 .or. ip >= 0 .or. is .ne. 0) goto 210
    is = 1
  elseif (ai .eq. '-') then
    if (id .ne. 0 .or. ip >= 0 .or. is .ne. 0) goto 210
    is = -1
  elseif (ai .eq. 'e' .or. ai .eq. 'E' .or. ai .eq. 'd' .or. ai .eq. 'D') then
    goto 100
  elseif (index (dig, ai) .eq. 0) then
    goto 210
  else
!    read (ai, '(f1.0)') bi
    bi = index (dig, ai) - 1
    if (inz > 0 .or. bi > 0.d0) then
      inz = 1
      id = id + 1
      call dsmuls (s1, 10., s0)
      f(1) = bi
      f(2) = 0.
      call dsdqc (bi, f)
      call dsadd (s0, f, s1)
    endif
  endif
enddo

100   continue
if (is .eq. -1) then
  s1(1) = - s1(1)
  s1(2) = - s1(2)
endif
k = i
if (ip == -1) ip = id
ie = 0
is = 0
ca = ' '

do i = k + 1, lnn
  ai = a(i:i)
  if (ai .eq. ' ') then
  elseif (ai .eq. '+') then
    if (ie .ne. 0 .or. is .ne. 0) goto 210
    is = 1
  elseif (ai .eq. '-') then
    if (ie .ne. 0 .or. is .ne. 0) goto 210
    is = -1
  elseif (index (dig, ai) .eq. 0) then
    goto 210
  else
    ie = ie + 1
    if (ie .gt. 3) goto 210
    ca(ie:ie) = ai
  endif
enddo

! read (ca, '(i4)') ie
ie = dsdigin (ca, 4)
if (is .eq. -1) ie = - ie
ie = ie + ip - id
s0(1) = 10.
s0(2) = 0.
call dsnpwr (s0, ie, s2)
call dsmul (s1, s2, b)
goto 220

210  write (6, 1)
1 format ('*** DSINPC: Syntax error in literal string.')
call dsabrt

220  return
end subroutine

subroutine dslog (a, b)

!   This computes the natural logarithm of the DS number A and returns the DS
!   result in B.

!   The Taylor series for Log converges much more slowly than that of Exp.
!   Thus this routine does not employ Taylor series, but instead computes
!   logarithms by solving Exp (b) = a using the following Newton iteration,
!   which converges to b:

!           x_{k+1} = x_k + [a - Exp (x_k)] / Exp (x_k)

!   These iterations are performed with a maximum precision level NW that
!   is dynamically changed, approximately doubling with each iteration.

implicit none
integer k
real*4 t1, t2
real*4 a(2), al2(2), b(2), s0(2), s1(2), s2(2)
save al2
data al2/6.93147182E-01, -1.90465421E-09/

if (a(1) .le. 0.) then
  write (6, 1)
1 format ('*** DSLOG: Argument is less than or equal to zero.')
  call dsabrt
  return
endif

!   Compute initial approximation of Log (A).

t1 = a(1)
t2 = log (t1)
b(1) = t2
b(2) = 0.

!   Perform the Newton-Raphson iteration described above.

do k = 1, 3
  call dsexp (b, s0)
  call dssub (a, s0, s1)
  call dsdiv (s1, s0, s2)
  call dsadd (b, s2, s1)
  b(1) = s1(1)
  b(2) = s1(2)
enddo

120  continue

return
end subroutine

subroutine dsqdc (a, b)

!   This converts the DS number A to DP.

implicit none
real*4 a(2), b

b = a(1)
return
end subroutine

subroutine dsqqc (a, b, c)

!   This converts DS numbers A and B to DSC form in C, i.e. C = A + B i.

implicit none
real*4 a(2), b(2), c(4)

c(1) = a(1)
c(2) = a(2)
c(3) = b(1)
c(4) = b(2)
return
end subroutine

subroutine dsmul (dsa, dsb, dsc)

!   This routine multiplies DS numbers DSA and DSB to yield the DS product DSC.

implicit none
real*4 dsa(2), dsb(2), dsc(2)
real*4 a1, a2, b1, b2, cona, conb, c11, c21, c2, e, split, t1, t2
parameter (split = 8193.)

!>
!   On systems with a fused multiply add, such as IBM systems, it is faster to
!   uncomment the next two lines and comment out the following lines until !>.
!   On other systems, do the opposite.

! c11 = dsa(1) * dsb(1)
! c21 = dsa(1) * dsb(1) - c11

!   This splits dsa(1) and dsb(1) into high-order and low-order words.

cona = dsa(1) * split
conb = dsb(1) * split
a1 = cona - (cona - dsa(1))
b1 = conb - (conb - dsb(1))
a2 = dsa(1) - a1
b2 = dsb(1) - b1

!   Multilply dsa(1) * dsb(1) using Dekker's method.

c11 = dsa(1) * dsb(1)
c21 = (((a1 * b1 - c11) + a1 * b2) + a2 * b1) + a2 * b2
!>
!   Compute dsa(1) * dsb(2) + dsa(2) * dsb(1) (only high-order word is needed).

c2 = dsa(1) * dsb(2) + dsa(2) * dsb(1)

!   Compute (c11, c21) + c2 using Knuth's trick, also adding low-order product.

t1 = c11 + c2
e = t1 - c11
t2 = ((c2 - e) + (c11 - (t1 - e))) + c21 + dsa(2) * dsb(2)

!   The result is t1 + t2, after normalization.

dsc(1) = t1 + t2
dsc(2) = t2 - (dsc(1) - t1)

return
end subroutine

subroutine dsmuls (dsa, db, dsc)

!   This routine multiplies the DS number DSA by the DP number DB to yield
!   the DS product DSC.

implicit none
real*4 dsa(2), db, dsc(2)
real*4 a1, a2, b1, b2, cona, conb, c11, c21, c2, e, split, t1, t2
parameter (split = 8193.)

!>
!   On systems with a fused multiply add, such as IBM systems, it is faster to
!   uncomment the next two lines and comment out the following lines until !>.
!   On other systems, do the opposite.

! c11 = dsa(1) * db
! c21 = dsa(1) * db - c11

!   This splits dsa(1) and db into high-order and low-order words.

cona = dsa(1) * split
conb = db * split
a1 = cona - (cona - dsa(1))
b1 = conb - (conb - db)
a2 = dsa(1) - a1
b2 = db - b1

!   Multilply dsa(1) * db using Dekker's method.

c11 = dsa(1) * db
c21 = (((a1 * b1 - c11) + a1 * b2) + a2 * b1) + a2 * b2
!>
!   Compute dsa(2) * db (only high-order word is needed).

c2 = dsa(2) * db

!   Compute (c11, c21) + c2 using Knuth's trick.

t1 = c11 + c2
e = t1 - c11
t2 = ((c2 - e) + (c11 - (t1 - e))) + c21

!   The result is t1 + t2, after normalization.

dsc(1) = t1 + t2
dsc(2) = t2 - (dsc(1) - t1)
return
end subroutine

subroutine dsmulss (da, db, dsc)

!   This subroutine computes dsc = da x db.

implicit none
real*4 a1, a2, b1, b2, cona, conb, da, db, dsc(2), split, s1, s2
parameter (split = 8193.)

!>
!   On systems with a fused multiply add, such as IBM systems, it is faster to
!   uncomment the next two lines and comment out the following lines until !>.
!   On other systems, do the opposite.

! s1 = da * db
! s2 = da * db - s1

!   This splits da and db into high-order and low-order words.

cona = da * split
conb = db * split
a1 = cona - (cona - da)
b1 = conb - (conb - db)
a2 = da - a1
b2 = db - b1

!   Multiply da * db using Dekker's method.

s1 = da * db
s2 = (((a1 * b1 - s1) + a1 * b2) + a2 * b1) + a2 * b2
!>
dsc(1) = s1
dsc(2) = s2

return
end subroutine

subroutine dsnint (a, b)

!   This sets B equal to the integer nearest to the DS number A.

implicit none
real*4 a(2), b(2), con(2), s0(2), t47, t23
parameter (t47 = 2. ** 47, t23 = 2. ** 23)
save con
data con / t47, t23/

!   Check if  A  is zero.

if (a(1) .eq. 0.)  then
  b(1) = 0.
  b(2) = 0.
  goto 120
endif

if (a(1) .ge. t47) then
  write (6, 1)
1 format ('*** DSINFR: Argument is too large.')
  call dsabrt
  return
endif

if (a(1) .gt. 0.) then
  call dsadd (a, con, s0)
  call dssub (s0, con, b)
else
  call dssub (a, con, s0)
  call dsadd (s0, con, b)
endif

120  continue
return
end subroutine

subroutine dsnpwr (a, n, b)

!   This computes the N-th power of the DS number A and returns the DS result
!   in B.  When N is zero, 1 is returned.  When N is negative, the reciprocal
!   of A ^ |N| is returned. 

!   This routine employs the binary method for exponentiation.

implicit none
integer j, kk, kn, l1, mn, n, na1, na2, nn
real*4 cl2, t1
parameter (cl2 = 1.4426950408889633)
real*4 a(2), b(2), s0(2), s1(2), s2(2), s3(2)

if (a(1) .eq. 0.) then
  if (n .ge. 0) then
    s2(1) = 0.
    s2(2) = 0.
    goto 120
  else
    write (6, 1)
1   format ('*** DSCPWR: Argument is zero and N is negative or zero.')
    call dsabrt
    return
  endif
endif

nn = abs (n)
if (nn .eq. 0) then
  s2(1) = 1.
  s2(2) = 0.
  goto 120
elseif (nn .eq. 1) then
  s2(1) = a(1)
  s2(2) = a(2)
  goto 110
elseif (nn .eq. 2) then
  call dsmul (a, a, s2)
  goto 110
endif

!   Determine the least integer MN such that 2 ^ MN .GT. NN.

t1 = nn
mn = cl2 * log (t1) + 1. + 1.e-7
s0(1) = a(1)
s0(2) = a(2)
s2(1) = 1.
s2(2) = 0.
kn = nn

!   Compute B ^ N using the binary rule for exponentiation.

do j = 1, mn
  kk = kn / 2
  if (kn .ne. 2 * kk) then
    call dsmul (s2, s0, s1)
    s2(1) = s1(1)
    s2(2) = s1(2)
  endif
  kn = kk
  if (j .lt. mn) then
    call dsmul (s0, s0, s1)
    s0(1) = s1(1)
    s0(2) = s1(2)
  endif
enddo

!   Compute reciprocal if N is negative.

110  continue

if (n .lt. 0) then
  s1(1) = 1.
  s1(2) = 0.
  call dsdiv (s1, s2, s0)
  s2(1) = s0(1)
  s2(2) = s0(2)
endif

120  continue

b(1) = s2(1)
b(2) = s2(2)
  
return
end subroutine

subroutine dsnrt (a, n, b)

!   This computes the N-th root of the DS number A and returns the DS result
!   in B.  N must be at least one.

!   This subroutine employs the following Newton-Raphson iteration, which
!   converges to A ^ (-1/N):

!    X_{k+1} = X_k + (X_k / N) * (1 - A * X_k^N)

!   The reciprocal of the final approximation to A ^ (-1/N) is the N-th root.

implicit none
integer i, k, n
real*4 t1, t2, tn
real*4 a(2), b(2), f1(2), s0(2), s1(2)

if (a(1) .eq. 0.) then
  b(1) = 0.
  b(2) = 0.
  goto 140
elseif (a(1) .lt. 0.) then
  write (6, 1)
1 format ('*** DSNRT: Argument is negative.')
  call dsabrt
  return
endif
if (n .le. 0) then
  write (6, 2) n
2 format ('*** DSNRT: Improper value of N',i10)
  call dsabrt
  return
endif

!   Handle cases N = 1 and 2.

if (n .eq. 1) then
  b(1) = a(1)
  b(2) = a(1)
  goto 140
elseif (n .eq. 2) then
  call dssqrt (a, b)
  goto 140
endif

f1(1) = 1.
f1(2) = 0.

!   Compute the initial approximation of A ^ (-1/N).

tn = n
t1 = a(1)
t2 = exp (- log (t1) / tn)
b(1) = t2
b(2) = 0.

!   Perform the Newton-Raphson iteration described above.

do k = 1, 3
  call dsnpwr (b, n, s0)
  call dsmul (a, s0, s1)
  call dssub (f1, s1, s0)
  call dsmul (b, s0, s1)
  call dsdivs (s1, tn, s0)
  call dsadd (b, s0, s1)
  b(1) = s1(1)
  b(2) = s1(2)
enddo

!   Take the reciprocal to give final result.

call dsdiv (f1, b, s1)
b(1) = s1(1)
b(2) = s1(2)

140  continue
return
end subroutine

subroutine dsout (iu, a)

!   This routine writes the DS number A on logical unit iu using a standard
!   E format, with lines 22 characters long.

implicit none
integer i, iu, ln
parameter (ln = 22)
character*22 cs
real*4 a(2)

call dsoutc (a, cs)
write (iu, '(a)') cs

return
end subroutine

subroutine dsoutc (a, b)

!   Converts the DS number A into character form in the CHARACTER*22 array B.
!   The format is analogous to the Fortran E format.

!   This routine is called by DSOUT, but it may be directly called by the user
!   if desired for custom output.

implicit none
integer i, ii, ix, ln, nx
parameter (ln = 22)
integer ib(ln)
real*4 t1
character*22 b
character*10 ca, digits
parameter (digits = '0123456789')
real*4 a(2), f(2), s0(2), s1(2)

f(1) = 10.
f(2) = 0.

do i = 1, ln
  ib(i) = 0
enddo

!   Determine exact power of ten for exponent.

if (a(1) .ne. 0.) then
  t1 = log10 (abs (a(1)))
  if (t1 .ge. 0.) then
    nx = t1
  else
    nx = t1 - 1.
  endif
  if (nx > 34) then
    do i = 1, ln
      b(i:i) = '*'
    enddo
    goto 200
  elseif (nx < -34) then
    b(1:1) = ' '
    b(2:2) = ' '
    b(3:3) = '0'
    b(4:4) = '.'
    do i = 5, ln - 4
      b(i:i) = '0'
    enddo
    b(ln-3:ln-3) = 'E'
    b(ln-2:ln-2) = ' '
    b(ln-1:ln-1) = ' '
    b(ln:ln) = '0'
    goto 200
  endif
  call dsnpwr (f, nx, s0)
  call dsdiv (a, s0, s1)
  if (s1(1) .lt. 0.) then
    s1(1) = - s1(1)
    s1(2) = - s1(2)
  endif

!   If we didn't quite get it exactly right, multiply or divide by 10 to fix.

i = 0

100 continue

  i = i + 1
  if (s1(1) .lt. 1.) then
    nx = nx - 1
    call dsmuls (s1, 10., s0)
    s1(1) = s0(1)
    s1(2) = s0(2)
    if (i <= 3) goto 100
  elseif (s1(1) .ge. 10.) then
    nx = nx + 1
    call dsdivs (s1, 10., s0)
    s1(1) = s0(1)
    s1(2) = s0(2)
    goto 100
  endif
else
  nx = 0
  s1(1) = 0.
  s1(2) = 0.
endif

!   Compute digits.

do i = 1, ln - 7
  ii = s1(1)
  ib(i) = ii
  f(1) = ii
  call dssub (s1, f, s0)
  call dsmuls (s0, 10., s1)
enddo

!   Fix negative digits.

do i = ln - 7, 2, -1
  if (ib(i) .lt. 0) then
    ib(i) = ib(i) + 10
    ib(i-1) = ib(i-1) - 1
  endif
enddo

if (ib(1) .lt. 0) then
  write (6, 1) 
1 format ('dsoutc: negative leading digit')
  call dsabrt
endif

!   Round.

if (ib(ln-7) .ge. 5) then
  ib(ln-8) = ib(ln-8) + 1

  do i = ln - 8, 2, -1
    if (ib(i) .eq. 10) then
      ib(i) = 0
      ib(i-1) = ib(i-1) + 1
    endif
  enddo

  if (ib(1) .eq. 10) then
    ib(1) = 1
    nx = nx + 1
  endif
endif

!   Insert digit characters in ib.

b(1:1) = ' '
b(2:2) = ' '
if (a(1) .ge. 0.) then
  b(3:3) = ' '
else
  b(3:3) = '-'
endif
ii = ib(1)
b(4:4) = digits(ii+1:ii+1)
b(5:5) = '.'
b(ln:ln) = ' '

do i = 2, ln - 8
  ii = ib(i)  
  b(i+4:i+4) = digits(ii+1:ii+1)
enddo

!   Insert exponent.

190  continue
! write (ca, '(i3)') nx
ca = dsdigout (real (nx), 3)
b(ln-3:ln-3) = 'E'

do i = 1, 3
  b(ln-3+i:ln-3+i) = ca(i:i)
enddo

200 continue

return
end subroutine

subroutine dspic (pi)

!   This returns Pi to quad precision.

implicit none
real*4 pi(2), pic(2)
save pic
data pic/ 3.14159274E+00, -8.74227766E-08/

pi(1) = pic(1)
pi(2) = pic(2)

return
end subroutine

subroutine dspoly (n, a, x0, x)

!   This finds the root x near x0 (input) for the nth-degree polynomial whose
!   coefficients are given in the n+1-long vector a.  It may be necessary to
!   adjust eps -- default value is 1e-14

implicit none
integer i, it, n
real*4  a(2,0:n), ad(2,0:n), t1(2), t2(2), t3(2), t4(2), t5(2), &
  x(2), x0(2), dt1, eps
parameter (eps = 1.e-14)

do i = 0, n - 1
  dt1 = i + 1
  call dsmuls (a(1,i+1), dt1, ad(1,i))
enddo

ad(1,n) = 0.
ad(2,n) = 0.
x(1) = x0(1)
x(2) = x0(2)

do it = 1, 20
  t1(1) = 0.
  t1(2) = 0.
  t2(1) = 0.
  t2(2) = 0.
  t3(1) = 1.
  t3(2) = 0.

  do i = 0, n
    call dsmul (a(1,i), t3, t4)
    call dsadd (t1, t4, t5)
    t1(1) = t5(1)
    t1(2) = t5(2)
    call dsmul (ad(1,i), t3, t4)
    call dsadd (t2, t4, t5)
    t2(1) = t5(1)
    t2(2) = t5(2)
    call dsmul (t3, x, t4)
    t3(1) = t4(1)
    t3(2) = t4(2)
  enddo

  call dsdiv (t1, t2, t3)
  call dssub (x, t3, t4)
  x(1) = t4(1)
  x(2) = t4(2)
  if (abs (t3(1)) .le. eps) goto 110
enddo

write (6, 1)
1 format ('dsroot: failed to converge.')
call dsabrt

110 continue

return
end subroutine

subroutine dssqrt (a, b)

!   This computes the square root of the DS number A and returns the DS result
!   in B.

!   This subroutine employs the following formula (due to Alan Karp):

!          Sqrt(A) = (A * X) + 0.5 * [A - (A * X)^2] * X  (approx.)

!   where X is a double precision approximation to the reciprocal square root,
!   and where the multiplications A * X and [] * X are performed with only
!   double precision.

implicit none
real*4 t1, t2, t3
real*4 a(2), b(2), f(2), s0(2), s1(2)

if (a(1) .eq. 0.) then
  b(1) = 0.
  b(2) = 0.
  goto 100
endif
t1 = 1. / sqrt (a(1))
t2 = a(1) * t1
call dsmulss (t2, t2, s0)
call dssub (a, s0, s1)
t3 = 0.5 * s1(1) * t1
s0(1) = t2
s0(2) = 0.
s1(1) = t3
s1(2) = 0.
call dsadd (s0, s1, b)

100 continue

return
end subroutine

subroutine dssub (dsa, dsb, dsc)

!   This subroutine computes dsc = dsa - dsb.

implicit none
real*4 dsa(2), dsb(2), dsc(2)
real*4 e, t1, t2

!   Compute dsa + dsb using Knuth's trick.

t1 = dsa(1) - dsb(1)
e = t1 - dsa(1)
t2 = ((-dsb(1) - e) + (dsa(1) - (t1 - e))) + dsa(2) - dsb(2)

!   The result is t1 + t2, after normalization.

dsc(1) = t1 + t2
dsc(2) = t2 - (dsc(1) - t1)
return
end subroutine

  real*4 function dsdigin (ca, n)
    implicit none
    real*4 d1
    character*(*), ca
    character*16 digits
    integer i, k, n
    parameter (digits = '0123456789')

    d1 = 0.

    do i = 1, n
      k = index (digits, ca(i:i)) - 1
      if (k < 0) then
        write (6, *) 'dsdigin: non-digit in character string'
      elseif (k <= 9) then
        d1 = 10. * d1 + k
      endif
    enddo

    dsdigin = d1
  end function

  character*16 function dsdigout (a, n)
    implicit none
    real*4 a, d1, d2
    character*16 ca, digits
    parameter (digits = '0123456789')
    integer i, is, k, n

    ca = ' '
    is = sign (1., a)
    d1 = abs (a)

    do i = n, 1, -1
      d2 = aint (d1 / 10.)
      k = 1. + (d1 - 10. * d2)
      d1 = d2
      ca(i:i) = digits(k:k)
      if (d1 == 0.) goto 100
    enddo

    i = 0

100 continue

    if (is < 0 .and. i > 1) then
      ca(i-1:i-1) = '-'
    elseif (i == 0 .or. is < 0 .and. i == 1) then
      ca = '****************'
    endif

    dsdigout = ca
    return
  end function

end module
